from abc import ABC , abstractmethod
import cloudsend.utils as cloudsendutils
import sys , json , os , time
import re
#import multiprocessing
import math , random
import cloudsend.combopt as combopt
from io import BytesIO
import csv , io
import pkg_resources
from cloudsend.core import *
from cloudsend.provider import CloudSendProvider , CloudSendProviderState , debug , PROVIDER_CONFIG
from cloudsend.config_state import ConfigManager , StateSerializer , STATE_FILE
from enum import IntFlag
from threading import current_thread
import shutil
import asyncio , asyncssh

random.seed()

SLEEP_PERIOD = 15

class CloudSendProviderStateWaitMode(IntFlag):
    NO_WAIT       = 0  # provider dont wait for state
    WAIT          = 1  # provider wait for state 
    WATCH         = 2  # provider watch out for state = wait + revive

class CloudSendFatProvider(CloudSendProvider,ABC):

    def __init__(self, conf=None):

        # this will call self._init
        CloudSendProvider.__init__(self,conf)

        # will be called by the parent
        #self._init(conf)

    def _init(self,conf):
        super()._init(conf)

        self._instances = []
        self._environments = []
        self._jobs = []
        self._run_sessions = []
        self._current_session = None        

        # load the config
        self._config_manager = ConfigManager(self,self._config,self._instances,self._environments,self._jobs)
        self._config_manager.load()

        # option
        self._mutualize_uploads = conf.get('mutualize_uploads',True)

        self._state_serializer = None
        if self._config.get('recover',False):
            # load the state (if existing) and set the recovery mode accordingly
            self._state_serializer = StateSerializer(self)
            self._state_serializer.load()

            consistency = self._state_serializer.check_consistency(self._state,self._instances,self._environments,self._jobs,self._run_sessions,self._current_session)
            if consistency:
                self.debug(1,"STATE RECOVERY: state is consistent with configuration - LOADING old state",color=bcolors.CVIOLET)
                self._recovery = True
                self._state , self._instances , self._environments , self._jobs , self._run_sessions , self._current_session = self._state_serializer.transfer()
                self.debug(2,self._instances)
                self.debug(2,self._environments)
                self.debug(2,self._jobs)
                self.debug(2,self._run_sessions)
                for job in self._jobs:
                    process = job.get_last_process()
                    self.debug(2,process)
            else:
                self._recovery = False
        else:
            self._recovery = False   

        #self._multiproc_man   = multiprocessing.Manager()
        #self._multiproc_lock  = self._multiproc_man.Lock()
        self._instances_watching = dict()
        self._instances_reviving = dict()

        # watch asyncio future
        self._watcher_task = None

        self._save_config() 

    async def _cfg_add_objects(self,conf,method):
        await method(conf)
        self._config_manager.load()    
        self.set_state( self._state & (CloudSendProviderState.NEW|CloudSendProviderState.STARTED))
        if self._state & CloudSendProviderState.STARTED:
            await self.start()

    async def cfg_add_instances(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_instances)

    async def cfg_add_environments(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_environments)

    async def cfg_add_jobs(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_jobs)

    async def cfg_add_config(self,conf):
        await self._cfg_add_objects(conf,super().cfg_add_config)

    async def cfg_reset(self):
        # remove the state file
        if self._state_serializer:
            self._state_serializer.reset()
        elif os.path.isfile(STATE_FILE): # let's always remove it
            try:
                os.remove(STATE_FILE)
            except:
                pass
        # provider reset: remove config file
        await super().cfg_reset()
        # re initialize objects
        self._init(self._config)

    def serialize_state(self):
        if self._config.get('recover',False):
            self._state_serializer.serialize(self._state,self._instances,self._environments,self._jobs,self._run_sessions,self._current_session)

    def set_state(self,value):
        self._state = value
        self.serialize_state()

    # def get_job(self,index):

    #     return self._jobs[index] 

    async def _assign(self):

        if self._recovery == True and self._state & CloudSendProviderState.ASSIGNED:
            assign_jobs = False
            for job in self._jobs:
                if not job.get_instance():
                    assign_jobs = True
                    break
            if not assign_jobs:
                self.debug(1,"STATE RECOVERY: skipping jobs allocation dues to reloaded state...",color=bcolors.CVIOLET)
                return 

        self.set_state(self._state & (CloudSendProviderState.ANY - CloudSendProviderState.ASSIGNED))

        assignation = self._config.get('job_assign','multi_knapsack')

        for instance in self._instances:
            instance.reset_jobs()
        
        # DUMMY algorithm 
        if assignation=='random':
            for job in self._jobs:
                if job.get_instance():
                    self.debug(2,"Job already has an instance",job)
                    continue
                
                instance = random.choice( self._instances )
                
                job.set_instance(instance)
                self.debug(1,"Assigned job " + str(job) )

        # knapsack , 2d packing , bin packing ...
        else: #if assignation is None or assignation=='multi_knapsack':

            combopt.multiple_knapsack_assignation(self._jobs,self._instances)   

        self.set_state( self._state | CloudSendProviderState.ASSIGNED )

        self.serialize_state()         
               
    async def _deploy_instance(self,instance,deploy_states,ssh_conn,ftp_client):

        homedir     = instance.get_home_dir()
        global_path = instance.get_global_dir()
        files_path  = instance.path_join(global_path,'files')
        ready_path  = instance.path_join(global_path,'ready')

        # last file uploaded ...
        re_upload  = await self._test_reupload(instance,ready_path,ssh_conn)

        #created = deploy_states[instance.get_name()].get('created')

        debug(2,"re_upload",re_upload)

        if re_upload:

            self.debug(2,"creating instance's directories ...")
            stdout0, stderr0 = await self._exec_command(ssh_conn,"mkdir -p "+global_path+" "+files_path+" && rm -f "+ready_path)
            self.debug(2,"directories created")

            self.debug(1,"uploading instance's files ... ")

            # upload the install file, the env file and the script file
            ftp_client = await ssh_conn.start_sftp_client()

            # change dir to global dir (should be done once)
            await ftp_client.chdir(global_path)
            for file in ['config.py','bootstrap.sh','run.sh','microrun.sh','state.sh','tail.sh','getpid.sh','reset.sh']:
                await self.sftp_put_remote_file(ftp_client,file) 

            self.debug(1,"Installing PyYAML for newly created instance ...")
            stdout, stderr = await self._exec_command(ssh_conn,"pip install pyyaml")
            self.debug(2,await stdout.read())
            self.debug(2, "Errors")
            self.debug(2,await stderr.read())

            commands = [ 
                # make bootstrap executable
                { 'cmd': "chmod +x "+instance.path_join(global_path,"*.sh"), 'out' : True },              
            ]

            await self._run_ssh_commands(instance,ssh_conn,commands)

            await self.sftp_put_string(ftp_client,'ready',"")

            self.debug(1,"files uploaded.")


        deploy_states[instance.get_name()] = { 'upload' : re_upload } 

    async def _deploy_environments(self,instance,deploy_states,ssh_conn,ftp_client):

        re_upload_inst = deploy_states[instance.get_name()]['upload']

        # scan the instances environment (those are set when assigning a job to an instance)
        #TODO: debug this
        # NOT SURE why we're missing an environment sometimes...
        bootstrap_command = ""

        for environment in instance.get_environments():
        #for environment in self._environments:

            # "deploy" the environment to the instance and get a DeployedEnvironment
            dpl_env  = environment.deploy(instance) 

            self.debug(3,dpl_env.json())

            ready_file = instance.path_join( dpl_env.get_path() , 'ready' )
            re_upload_env = await self._test_reupload(instance,ready_file, ssh_conn)

            re_upload_env_mamba  = False
            re_upload_env_pip    = False
            re_upload_env_aptget = False

            if not re_upload_env:
                if dpl_env.get_config('env_conda') is not None:
                    mamba_test = instance.path_join( instance.get_home_dir() , 'micromamba' , 'envs' , dpl_env.get_name_with_hash() )
                    re_upload_env_mamba = await self._test_reupload(instance,mamba_test, ssh_conn,False)
                    re_upload_env = re_upload_env or re_upload_env_mamba
                if dpl_env.get_config('env_pypi') is not None and dpl_env.get_config('env_conda') is None:
                    venv_test = instance.path_join( instance.get_home_dir() , '.' + dpl_env.get_name_with_hash() )
                    re_upload_env_pip = await self._test_reupload(instance,venv_test, ssh_conn, False)
                    re_upload_env = re_upload_env or re_upload_env_pip
                # TODO: have an aptget install TEST
                #if dpl_env.get_config('env_aptget') is not None:
                #    re_upload_env_aptget = True
                #    re_upload_env = True

            debug(2,"re_upload_instance",re_upload_inst,"re_upload_env",re_upload_env,"re_upload_env_mamba",re_upload_env_mamba,"re_upload_env_pip",re_upload_env_pip,"re_upload_env_aptget",re_upload_env_aptget,"ENV",dpl_env.get_name_with_hash())

            re_upload = re_upload_env #or re_upload_inst

            deploy_states[instance.get_name()][environment.get_name_with_hash()] = { 'upload' : re_upload }

            if re_upload:
                files_path = dpl_env.get_path()
                global_path = instance.get_global_dir() 

                self.debug(2,"creating environment directories ...")
                stdout0, stderr0 = await self._exec_command(ssh_conn,"mkdir -p "+files_path+" && rm -f "+ready_file)
                self.debug(2,"STDOUT for mkdir -p ",files_path,"...",await stdout0.read())
                self.debug(2,"STDERR for mkdir -p ",files_path,"...",await stderr0.read())
                self.debug(2,"directories created")

                self.debug(1,"uploading files ... ")

                # upload the install file, the env file and the script file
                # change to env dir
                await ftp_client.chdir(dpl_env.get_path())
                await self.sftp_put_string(ftp_client,'config.json',dpl_env.json())

                self.debug(1,"uploaded.")        

                print_deploy = self._config.get('print_deploy',False) == True

                config_py = instance.path_join( global_path , 'config.py' )

                commands = [
                    # recreate pip+conda files according to config
                    { 'cmd': "cd " + files_path + " && python3 "+config_py , 'out' : True },
                    # setup envs according to current config files state
                    # FIX: mamba is not handling well concurrency
                    # we run the mamba installs sequentially below
                    #{ 'cmd': instance.path_join( global_path , 'bootstrap.sh' ) + " \"" + dpl_env.get_name_with_hash() + "\" " + ("1" if self._config['dev'] else "0") , 'out': print_deploy , 'output': instance.path_join( dpl_env.get_path() , 'bootstrap._') },  
                ]

                bootstrap_command = bootstrap_command + (" ; " if bootstrap_command else "") + instance.path_join( global_path , 'bootstrap.sh' ) + " \"" + dpl_env.get_name_with_hash() + "\" " + ("1" if self._config['dev'] else "0")

                await self._run_ssh_commands(instance,ssh_conn,commands)
                
        if bootstrap_command:
            gbl_dir = instance.get_global_dir()
            await ftp_client.chdir(gbl_dir)
            await self.sftp_put_string(ftp_client,'generate_envs.sh',bootstrap_command)
            generate_sh = instance.path_join( gbl_dir , 'generate_envs.sh' ) 
            bootstrap_log = instance.path_join( gbl_dir , 'bootstrap.log' )
            commands = [
                {'cmd': 'chmod +x ' + generate_sh , 'out':True}, # import to wait for this to be done !
                {'cmd': generate_sh , 'out':print_deploy, 'output': bootstrap_log }
            ]
            await self._run_ssh_commands(instance,ssh_conn,commands)
        

    async def _deploy_jobs(self,instance,deploy_states,ssh_conn,ftp_client):

        file_uploaded = dict()

        # scan the instances environment (those are set when assigning a job to an instance)
        for job in instance.get_jobs():
            env      = job.get_env()        # get its environment
            dpl_env  = env.deploy(instance) # "deploy" the environment to the instance and get a DeployedEnvironment
            dpl_job  = job.deploy(dpl_env,False) # do not add this dpl_job permanently (only use for utility here)

            input_files = []
            if job.get_config('upload_files'):
                upload_files = job.get_config('upload_files')
                if isinstance(upload_files,str):
                    input_files.append(upload_files)
                else:
                    for upf in upload_files:
                        input_files.append(upf)
            if job.get_config('input_file'):
                input_files.append(job.get_config('input_file'))
            
            mkdir_cmd = ""
            for in_file in input_files:
                local_path , rel_path , abs_path , rel_remote_path , external = self._resolve_dpl_job_paths(in_file,dpl_job)
                dirname = instance.path_dirname(abs_path)
                if dirname:
                    mkdir_cmd = mkdir_cmd + (" && " if mkdir_cmd else "") + "mkdir -p " + dirname 

            self.debug(2,"creating job directories ...")
            stdout0, stderr0 = await self._exec_command(ssh_conn,"mkdir -p "+dpl_job.get_path())
            if mkdir_cmd != "":
                stdout0, stderr0 = await self._exec_command(ssh_conn,mkdir_cmd)
            self.debug(2,"directories created")

            re_upload_env = deploy_states[instance.get_name()][env.get_name_with_hash()]['upload']

            ready_file = instance.path_join( dpl_job.get_path() , 'ready' )
            re_upload = await self._test_reupload(instance,ready_file, ssh_conn)

            self.debug(2,"re_upload_env",re_upload_env,"re_upload",re_upload)

            if re_upload: #or re_upload_env:

                stdout0, stderr0 = await self._exec_command(ssh_conn,"rm -f "+ready_file)

                self.debug(1,"uploading job files ... ",dpl_job.get_hash())

                global_path = instance.get_global_dir()
                files_dir = instance.path_join( global_path , 'files' )

                # change to job hash dir
                await ftp_client.chdir(dpl_job.get_path())
                if job.get_config('run_script'):
                    script_args = job.get_config('run_script').split()
                    script_file = script_args[0]
                    filename = os.path.basename(script_file)
                    try:
                        await ftp_client.put(os.path.abspath(script_file),filename)
                    except:
                        self.debug(1,"You defined a script that is not available",job.get_config('run_script'))

                # NEW ! WE now go to 
                # COMMENT THIS IF YOU WANT TO PUT FILES in relation to the jobs' dir
                if self._mutualize_uploads:
                    await ftp_client.chdir(files_dir)

                if job.get_config('upload_files'):
                    files = job.get_config('upload_files')
                    if isinstance( files,str):
                        files = [ files ] 
                    for upfile in files:
                        local_path , rel_path , abs_path , rel_remote_path , external = self._resolve_dpl_job_paths(upfile,dpl_job)
                                
                        # check if the remote path has already been uploaded ...
                        if abs_path in file_uploaded:
                            self.debug(2,"skipping upload of file",upfile,"for job#",job.get_rank(),"(file has already been uploaded)")
                            continue
                        file_uploaded[abs_path] = True
                        
                        try:
                            try:
                                await ftp_client.put(local_path,rel_remote_path) #os.path.basename(upfile))
                            except Exception as e:
                                self.debug(1,"You defined an upload file that is not available",upfile)
                                self.debug(1,e)
                        except Exception as e:
                            print("Error while uploading file",upfile)
                            print(e)
                if job.get_config('input_file'):

                    local_path , rel_path , abs_path , rel_remote_path , external = self._resolve_dpl_job_paths(job.get_config('input_file'),dpl_job)

                    if abs_path in file_uploaded:
                        self.debug(2,"skipping upload of file",upfile,"for job#",job.get_rank(),"(file has already been uploaded)")
                    else:
                        file_uploaded[abs_path] = True
                        #filename = os.path.basename(job.get_config('input_file'))
                        try:
                            await ftp_client.put(local_path,rel_remote_path) #job.get_config('input_file')) #filename)
                        except:
                            self.debug(1,"You defined an input file that is not available:",job.get_config('input_file'))
                
                # used to check if everything is uploaded
                await ftp_client.chdir(dpl_job.get_path())
                await self.sftp_put_string(ftp_client, 'ready', "")

                self.debug(1,"uploaded.",dpl_job.get_hash())

    async def _deploy_all(self,instance):

        deploy_states = dict()

        deploy_states[instance.get_name()] = { }

        attempts = 0 

        for job in instance.get_jobs():
            self.debug(3,"PROCESS in deploy_all",job.get_last_process())

        while attempts < 5:

            if attempts!=0:
                self.debug(1,"Trying again ...")

            instanceid , ssh_conn , ftp_client = await self._wait_and_connect(instance)

            if ssh_conn is None:
                self.debug(1,"ERROR: could not deploy instance",instance,color=bcolors.FAIL)
                return

            try :

                self.debug(1,"-- deploy instances --")

                await self._deploy_instance(instance,deploy_states,ssh_conn,ftp_client)

                self.debug(1,"-- deploy environments --")

                await self._deploy_environments(instance,deploy_states,ssh_conn,ftp_client)

                self.debug(1,"-- deploy jobs --")

                await self._deploy_jobs(instance,deploy_states,ssh_conn,ftp_client) 

                #ftp_client.close()
                ssh_conn.close()

                break

            except FileNotFoundError as fne:
                self.debug(1,fne)
                self.debug(1,"Filename = ",fne.filename)
                self.debug(1,"Error while deploying")
                #sys.exit() # this only kills the thread
                #os.kill(os.getpid(), signal.SIGINT)
                os._exit(1)
                raise fne

            except Exception as e:
                self.debug(1,e)
                self.debug(1,"Error while deploying")
                #sys.exit() # this only kills the thread
                #os.kill(os.getpid(), signal.SIGINT)
                os._exit(1)
                raise e
            
            attempts = attempts + 1

    # use this to make sure we're not blocking in the generator loop below ...
    # (allow full multithreading)
    async def _start_and_update_and_reset_instance(self,instance,reset):
        self._start_and_update_instance(instance)
        if reset:
           await self.reset_instance(instance)

    

    async def start(self,reset=False):

        if not reset and self._state & CloudSendProviderState.STARTED & self._recovery:
            self.debug(1,"STATE RECOVERY: we're not skipping start() in order to update instance information",color=bcolors.CVIOLET)

        self.set_state( self._state & (CloudSendProviderState.ANY - CloudSendProviderState.STARTED) )

        self._instances_states = dict() 
        self.debug(3,"Starting ...")
        jobs_wait = [ ] 
        for instance in self._instances:
            jobs_wait.append( self._start_and_update_and_reset_instance(instance,reset) )
        await asyncio.gather( * jobs_wait )

        for instance in self._instances:
            if instance.is_invalid():
                self.debug(1,"ERROR: Your configuration is causing an instance to not be created. Please fix.",inst.get_config_DIRTY(),color=bcolors.FAIL)
                sys.exit()

        self.set_state( self._state | CloudSendProviderState.STARTED )

    async def reset_instance(self,instance):
        self.debug(1,'RESETTING instance',instance.get_name())
        instanceid, ssh_conn , ftp_client = await self._wait_and_connect(instance)
        if ssh_conn is not None:
            await self.sftp_put_remote_file(ftp_client,'reset.sh')
            reset_file = instance.path_join( instance.get_home_dir() , 'reset.sh' )
            commands = [
                { 'cmd' : 'chmod +x '+reset_file+' && ' + reset_file , 'out' : True }
            ]
            await self._run_ssh_commands(instance,ssh_conn,commands)
            #ftp_client.close()
            ssh_conn.close()
        self.debug(1,'RESETTING done')    


    async def hard_reset_instance(instance):        
        await super().hard_reset_instance(instance)
        await self._deploy_all(instance)

    # GREAT summary
    # https://www.integralist.co.uk/posts/python-asyncio/

    # deploys:
    # - instances files
    # - environments files
    # - shared script files, uploads, inputs ...
    async def deploy(self):

        if not self._state & CloudSendProviderState.STARTED:
            self.debug(1,"Not ready. Call 'start' first",color=bcolors.WARNING)
            return

        if self._state & CloudSendProviderState.DEPLOYED:
            if self._recovery == True:
                self.debug(1,"STATE RECOVERY: skipping deploy due to reloaded state ...",color=bcolors.CVIOLET)
            else:
                self.debug(1,"skipping deploy (already deployed) ...",color=bcolors.OKCYAN)
            return 

        self.set_state( self._state & (CloudSendProviderState.ANY - CloudSendProviderState.DEPLOYED) )

        await self._assign()

        clients = {} 

        for job in self._jobs:
            self.debug(3,"PROCESS in deploy",job.get_last_process())
        
        jobs = []
        for instance in self._instances:
            jobs.append( self._deploy_all(instance) )
        await asyncio.gather( *jobs )

        self.set_state( self._state | CloudSendProviderState.DEPLOYED )

    async def _revive(self,run_session,instance):
        self.debug(1,"REVIVING instance",instance,color=bcolors.OKCYAN)

        jobs_can_be_saved = False
        if instance.get_state()==CloudSendInstanceState.STOPPED or instance.get_state()==CloudSendInstanceState.STOPPING:
            self.debug(1,"Instance is stopping|stopped, we just have to re-start the instance",instance,color=bcolors.OKCYAN)
            jobs_can_be_saved = True
            #self.start_instance(instance)
            await self._wait_for_instance(instance)
        else:
            # try restarting it
            self._start_and_update_instance(instance)
            # wait for it
            #no need - deploy is doing this already
            #self._wait_for_instance(instance)
            # re-deploy it
            await self._deploy_all(instance)
        if run_session: 
            await self._run(run_session,instance,jobs_can_be_saved) #will run the jobs for this instance

    def _mark_aborted(self,run_session,instance,state_mask):
        if run_session:
            run_session.mark_aborted(instance,state_mask) # mark ABORTED 

    async def get_log(self,process,ssh_conn):
        uid   = process.get_uid()
        job   = process.get_job() # dpl job
        env   = job.get_env() # dpl env
        jhash = job.get_hash()
        path  = env.get_path()
        instance = job.get_instance()
        try:
            run_path = instance.path_join( path , jhash , uid )
            run_log1 = instance.path_join( run_path , 'run-'+uid+'.log' )
            run_log2 = instance.path_join( run_path , 'run.log' )
            stdout , stderr = await self._exec_command(ssh_conn,"cat "+run_log1+' '+run_log2)
            log = await stdout.read()
            return log
        except:
            return None

        # we may not have to do anything:
        # we recovered a state at startup
        # let's make sure the state has "advanced":
        # 1) wait jobs have not aborted
        # 2) queue jobs have not aborted
        # 3) IDLE jobs are ...
        # ...
        # 4) basically the state is > old state ...?

    async def _check_run_state(self,run_session,instance,ssh_conn):
        if instance.get_name() in self._instances_states and self._instances_states[instance.get_name()]['changed']==True:
           self.debug(1,"Instance has changed! States of old jobs should return UNKNOWN and a new batch of jobs will be started",color=bcolors.WARNING)
           #let's just let the following logic do its job ... JOB CENTRIC 
           #return True , False
        
        # we're gonna compare the processes' states of the current 'run_session'
        active_processes_old = run_session.get_active_processes(instance)
        active_processes_new = []
        for p in active_processes_old:
            active_processes_new.append(copy.copy(p))
        do_run = True
        
        # fetch the states for active_processes_new list 
        fetched , ssh_conn = await self.__fetch_states_internal(run_session,instance,active_processes_new,True,ssh_conn)
        do_run = False
        all_done = True
        for process_old in active_processes_old:
            uid = process_old.get_uid()
            process_new = None
            # find the new process equivalent
            for p in active_processes_new:
                if p.get_uid() == uid:
                    process_new = p
                    break
            if process_new is None:
                self.debug(1,"Internal ERROR: process_new is NONE! Should not happen!",color=bcolors.FAIL)
                sys.exit(2)

            state_new = process_new.get_state()
            state_old = process_old.get_state()

            # we found one process that hasn't advanced ...
            # let's just run the jobs ...
            #TODO: improve precision of recovery
            self.debug(2,state_old.name,"vs",state_new.name)
            if state_new!=CloudSendProcessState.DONE and (state_new < state_old or (state_new == CloudSendProcessState.ABORTED or state_old == CloudSendProcessState.ABORTED or state_new == CloudSendProcessState.UNKNOWN or state_old == CloudSendProcessState.UNKNOWN)):
                self.debug(1,"We will run the following job because of an unsatisfying state. Job#",process_new.get_job().get_rank(),"=",process_new.get_state())
                do_run = True
                # do not break cause we want to check all_done properly!
                #break
            all_done = all_done and state_new == CloudSendProcessState.DONE

        if all_done:
            do_run = False

        # it is now time to update the old (memory connected) processes 
        # we've retrieved what we needed and we want to new state to be corrected for future prints
        # this should likely set the states to UNKNOWN if this is a new instance
        # (note: an ABORTED state will not switch to UNKNOWN state - in order to keep the most information)
        fetched , ssh_conn = await self.__fetch_states_internal(run_session,instance,active_processes_old,True,ssh_conn)

        if do_run:
            return True , False , ssh_conn
        else:
            # we return the old ones because those are the ones linked with the memory 
            # the other ones have been separated with copy.copy ...
            return False , all_done , ssh_conn

    async def _ensure_watch(self):
        if not self.is_watching():
            self.debug(1,"Started watching ...")
            await self._watch()



    async def _run_jobs_for_instance(self,run_session,batch,instance,except_done) :

        # we're not coming from revive but we've recovered a state ...
        # if except_done == False and self._recovery == True:
        #     self.debug(1,"STATE RECOVERY: found serialized state: we will not restart jobs that have completed",color=bcolors.CVIOLET)
        #     except_done = True

        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(instance)
        if ssh_conn is None:
            ssh_conn = await self._handle_instance_disconnect(run_session,instance,True,"could not run jobs for instance")
            if ssh_conn is None:
                self.debug(1,"Could not run jobs for instance",instance,color=bcolors.FAIL)
                return 
        
        #if self._recovery:
        if except_done and self._recovery:
            do_run , all_done , ssh_conn = await self._check_run_state(run_session,instance,ssh_conn)
            if not do_run:
                if all_done:
                    self.debug(1,"STATE RECOVERY: skipping run_jobs because the jobs have completed since we left them :) @",instance.get_name(),color=bcolors.CVIOLET)
                else:
                    self.debug(1,"STATE RECOVERY: skipping run_jobs because the jobs have advanced as we left them :) @",instance.get_name(),color=bcolors.CVIOLET)
                return

        global_path = instance.get_global_dir()

        cmd_run     = ""
        cmd_run_pre = "" 
        cmd_pid     = ""

        for job in instance.get_jobs():

            if except_done and job.has_completed():
                continue

            # should literally never happen
            if not job.get_instance():
                debug(1,"The job",job,"has not been assigned to an instance!",color=bcolors.FAIL)
                continue
            
            # FOR NOW
            env      = job.get_env()        # get its environment
            # "deploy" the environment to the instance and get a DeployedEnvironment 
            # note: this has already been done in deploy but it doesnt matter ... 
            #       we dont store the deployed environments, and we base everything on remote state ...
            # NOTE: this could change and we store every thing in memory
            #       but this makes it less robust to states changes (especially remote....)
            dpl_env  = env.deploy(instance)
            # deploy the job (adds)
            dpl_job  = job.deploy(dpl_env)
            # the batch creates the process
            process  = batch.create_process(dpl_job)

            if process is None:
                sys.exit(1)
            
            self.debug(2,process) 

            files_path  = dpl_env.get_path()

            run_path    = instance.path_join( dpl_job.get_path() , process.get_uid() )
            pid_file    = instance.path_join( run_path , 'pid' )
            state_file  = instance.path_join( run_path , 'state' )

            is_first = (cmd_run_pre=="")
            cmd_run_pre = cmd_run_pre + "rm -f " + pid_file + " && "
            cmd_run_pre = cmd_run_pre + "mkdir -p " + run_path + " && "
            if is_first: # first sequential script is waiting for bootstrap to be done by default
                cmd_run_pre = cmd_run_pre + "echo 'wait' > " + state_file + "\n"
            else: # all other scripts will be queued
                cmd_run_pre = cmd_run_pre + "echo 'queue' > " + state_file + "\n"

            uid = process.get_uid()

            ln_command = self._get_ln_command(dpl_job,uid)
            self.debug(2,ln_command)
            if ln_command != "":
                cmd_run_pre = cmd_run_pre + ln_command + "\n"

            run_sh  = instance.path_join( global_path , 'run.sh' )
            run_log = instance.path_join( run_path , 'run-'+uid+'.log' )
            pid_sh  = instance.path_join( global_path , 'getpid.sh' )
            cmd_run = cmd_run + run_sh+" \"" + dpl_env.get_name_with_hash() + "\" \""+dpl_job.get_command()+"\" " + job.get_config('input_file') + " " + job.get_config('output_file') + " " + job.get_hash()+" "+uid+">"+run_log+" 2>&1"
            cmd_run = cmd_run + "\n"
            cmd_pid = cmd_pid + pid_sh + " \"" + pid_file + "\"\n"

        tryagain = True

        while tryagain:

            batch_run_file = instance.path_join( global_path , 'batch_run-'+batch.get_uid()+'.sh')
            batch_pid_file = instance.path_join( global_path , 'batch_pid-'+batch.get_uid()+'.sh')
            try:
                ftp_client = await ssh_conn.start_sftp_client()
                await ftp_client.chdir(global_path)
                await self.sftp_put_string(ftp_client,batch_run_file,cmd_run_pre+cmd_run)
                await self.sftp_put_string(ftp_client, batch_pid_file,cmd_pid)
                # run
                commands = [ 
                    { 'cmd': "chmod +x "+batch_run_file+" "+batch_pid_file, 'out' : True } ,  # important to wait for it >> True !!!
                    # execute main script (spawn) (this will wait for bootstraping)
                    { 'cmd': batch_run_file , 'out' : False } 
                ]
                
                await self._run_ssh_commands(instance,ssh_conn,commands)
                tryagain = False
            except Exception as e:
                self.debug(1,e)
                self.debug(1,"ERROR: the instance is unreachable while sending batch",instance,color=bcolors.FAIL)
                ssh_conn = await self._handle_instance_disconnect(run_session,instance,True,"could not run jobs for instance")
                if ssh_conn is None:
                    return 
                tryagain = True

        ssh_conn.close()

        self.serialize_state()

    # entry point ...
    async def run(self):

        # we were actually still running processes
        # >> let's use the current session that has been loaded
        if self._recovery and self._state & CloudSendProviderState.RUNNING:
            self.debug(1,"STATE RECOVERY: continuing run job ...",color=bcolors.CVIOLET)
            await self._run(self._current_session,None,True,False)
        else:

            if not self._state & CloudSendProviderState.DEPLOYED:
                method_to_call = self._get_method_to_call()
                self.debug(1,"Not ready for run. Call '"+method_to_call+"' first.",color=bcolors.WARNING)
                return

            if self._recovery and self._state & CloudSendProviderState.IDLE:
                self.debug(1,"STATE RECOVERY: jobs have been ran already - about the run again",color=bcolors.CVIOLET)

            # we're gonna create a new run session ... deativate entirely the old one
            if self._current_session:
                self._current_session.deactivate() 

            # create the new session
            number = len(self._run_sessions)
            self._current_session = CloudSendRunSession(number)
            self._run_sessions.append(self._current_session)

            # run the new session
            await self._run(self._current_session)

    async def _run(self,run_session,instance_filter=None,except_done=False,do_init=True):

        # update the Provider state
        self.set_state( self._state & (CloudSendProviderState.ANY - CloudSendProviderState.RUNNING - CloudSendProviderState.IDLE) )

        if instance_filter:
            instances = [ instance_filter ] 
        else:
            instances = self._instances

        if do_init:
            # we're about the re-run the processes for 'instances'
            # we need to de-activate the processes of the batch on those instances...
            for instance in instances:
                run_session.deactivate(instance)

        # a batch is accross instances, create it now
        # IMPORTANT: create it after de-activation !
        batch = run_session.create_batch()

        # run the jobs on each instances
        jobs = []
        for instance in instances:
            jobs.append( self._run_jobs_for_instance(run_session,batch,instance,except_done) ) 
        await asyncio.gather( *jobs )

        # update the Provider state
        self.set_state( self._state | CloudSendProviderState.RUNNING )

        # Now we can start wathing !
        # NOTE: do not ensure watch at the beginning of run...
        #       not sure why but this causes the GatheringFuture to fail
        await self._ensure_watch()


    def print_jobs_summary(self,run_session=None,instance=None):
        if not self._state & CloudSendProviderState.STARTED:
            self.debug(1,"Not ready to get print jobs. Call 'start' first",color=bcolors.WARNING)
            return

        jobs = instance.get_jobs() if instance is not None else self._jobs
        # the lock is to make sure the prints are not scrambled 
        # when coming back from the instance at the same time ...
        #with self._multiproc_lock:
        self.debug(1,"\n----------------------------------------------------------------------------------------------------------------------------------------------------------")
        if instance:
            self.debug(1,instance.get_name(),instance.get_ip_addr())
        for i,job in enumerate(jobs):
            self.debug(1,"\nJob",job.get_rank(),"=",job.str_simple() if instance else job)
            dpl_jobs = job.get_deployed_jobs()
            for dpl_job in dpl_jobs:
                for process in dpl_job.get_processes():
                    if run_session:
                        session = process.get_batch().get_session()
                        if run_session != session:
                            continue
                    self.debug(1,"|_",process.str_simple())

    async def print_aborted_logs(self,run_session=None,instance=None):
        if not self._state & CloudSendProviderState.STARTED:
            self.debug(1,"Not ready to get logs. Call 'start' first",color=bcolors.WARNING)
            return

        instances = self._instances if instance is None else [ instance ]
        for _instance in instances:
            ssh_conn   = None
            ftp_client = None
            for job in _instance.get_jobs():
                process = job.get_last_process()
                if process is None: # should not really happened but it did while writing code
                    continue
                if process.get_state() == CloudSendProcessState.ABORTED:
                    if run_session:
                        session = process.get_batch().get_session()
                        if run_session != session:
                            continue
                    if not ssh_conn:
                        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(_instance)
                    log = await self.get_log(process,ssh_conn)
                    self.debug(1,"----------------------------------------------------------------------",color=bcolors.WARNING)
                    self.debug(1,"Job #",job.get_rank(),"has ABORTED with errors:",color=bcolors.WARNING)
                    self.debug(1,log,color=bcolors.WARNING)
                    self.debug(1,process,color=bcolors.WARNING)
            if ssh_conn:
                # ftp_client.close()
                ssh_conn.close()   

    async def print_objects(self):
        self.debug(1,"STATE =",self._state)
        self.debug(1,"\nINSTANCES: --------------------")
        for instance in self._instances:
            self.debug(1,instance)
        self.debug(1,"\nENVIRONMENTS: -----------------")
        for env in self._environments:
            self.debug(1,env)
        self.debug(1,"\nJOBS: -------------------------")
        for job in self._jobs:
            self.debug(1,job)

    def _get_method_to_call(self):
        if not self._state   & CloudSendProviderState.STARTED:
            return 'start'
        elif not self._state & CloudSendProviderState.DEPLOYED:
            return 'deploy'
        elif not self._state & CloudSendProviderState.RUNNING:
            return 'run'
        elif not self._state & CloudSendProviderState.ASSIGNED:
            return 'assign'
        else:
            return ''        

    async def fetch_results(self,out_dir,run_session=None):

        if not run_session:
            run_session = self._current_session

        if not run_session:
            self.debug(1,"No result to fetch",color=bcolors.WARNING)
            return 

        #processes = self.get_last_processes()
        processes = run_session.get_ran_processes()

        clients   = dict()

        try:
            shutil.rmtree(out_dir, ignore_errors=True)
        except:
            pass
        try:
            os.makedirs(out_dir)
        except:
            pass

        for process in processes:

            if process.get_state() != CloudSendProcessState.DONE:
                self.debug(2,"Skipping process import",process.get_uid(),process.get_state())
            
            dpl_job  = process.get_job()
            rank     = dpl_job.get_rank()
            instance = dpl_job.get_instance()

            if not instance.get_name() in clients:
                clients[instance.get_name()] = dict()
                instanceid , ssh_conn , ftp_client = await self._wait_and_connect(instance)
                if ssh_conn is not None:
                    #ftp_client = await ssh_conn.start_sftp_client()
                    clients[instance.get_name()] = { 'ssh': ssh_conn , 'ftp' : ftp_client}
                else:
                    self.debug(1,"Skipping instance",instance.get_name(),"(unreachable)",color=bcolors.WARNING)
            else:
                ssh_conn = clients[instance.get_name()]['ssh']
                ftp_client = clients[instance.get_name()]['ftp']
            
            out_file = dpl_job.get_config('output_file') # this file is written for the local machine
            remote_file_path = instance.path_join( process.get_path() , out_file )
            directory = instance.path_dirname( remote_file_path )
            filename  = instance.path_basename( remote_file_path )
            
            file_name , file_extension = os.path.splitext(out_file)
            file_name = file_name.replace(os.sep,'_')
            #ftp_client.chdir(directory)
            local_path = os.path.join(out_dir,'job_'+str(rank).zfill(3)+'_'+file_name+file_extension)
            await ftp_client.chdir( directory )
            try:
                await ftp_client.get( filename , local_path )   
            except asyncssh.sftp.SFTPNoSuchFile:
                self.debug(1,"No file for process",process.get_uid(),"job#",rank,directory,filename,local_path)
                pass
        
        for client in clients.values():
            #client['ftp'].close()
            client['ssh'].close()

    def _get_ln_command(self,dpl_job,uid):
        files_to_ln = []
        upload_files = dpl_job.get_config('upload_files')
        lnstr = ""
        instance = dpl_job.get_instance()
        if upload_files:
            for up_file in upload_files:
                files_to_ln.append(up_file)
        if dpl_job.get_config('input_file'):
            files_to_ln.append(dpl_job.get_config('input_file'))
        for upfile in files_to_ln:
            local_path , rel_path , abs_path , rel_remote_path , external = self._resolve_dpl_job_paths(upfile,dpl_job)
            filename    = instance.path_basename(abs_path)
            filedir_abs = instance.path_dirname(abs_path)
            filedir_rel = instance.path_dirname(rel_path)
            if filedir_rel and filedir_rel != instance.path_sep() :
                fulldir   = instance.path_join(dpl_job.get_path() , uid , filedir_rel)
                uploaddir = instance.path_join(dpl_job.get_path() , filedir_rel )
                full_file_path = instance.path_join( fulldir , filename )
                lnstr = lnstr + (" && " if lnstr else "") + "mkdir -p " + fulldir + " && ln -sf " + abs_path + " " +  full_file_path
            else:
                fulldir   = instance.path_join( dpl_job.get_path() , uid )
                fulldir2  = dpl_job.get_path() # let's also put symbolic links by the file itself ... 
                uploaddir = dpl_job.get_path()
                full_file_path  = instance.path_join( fulldir  , filename )
                full_file_path2 = instance.path_join( fulldir2 , filename )
                lnstr = lnstr + (" && " if lnstr else "") + "ln -sf " + abs_path + " " + full_file_path + " && ln -sf " + abs_path + " " + full_file_path2

        return lnstr

    async def _handle_instance_disconnect(self,run_session,instance,do_revive,msg):

        # let the priority to the watching/reviving process (should only be one)
        # this is really just in case ...
        # this is especially for the test below: instance.get_state() == CloudSendInstanceState.RUNNING:
        # we want to make sure that no other process may have re-started (quickly) the instance
        # NOTE: this was relevant when we had *concurrent* and *similar* watch and wait processes
        # not the case anymore.... wait is now leveraging watch
        if not do_revive:
            await asyncio.sleep(SLEEP_PERIOD)
            while self._instances_reviving[instance.get_name()]:
                self.debug(1,"waiting for reviving instance (2)",instance)
                await asyncio.sleep(SLEEP_PERIOD)
        else:
            self._instances_reviving[instance.get_name()] = True

        if self._instances_watching.get(instance.get_name(),False) == False:
            self.debug(1,"We have stopped watching the instance - We won't try to reconnect",color=bcolors.WARNING)
            return None 
            
        try:
            # check the status on the instance with AWS
            self.update_instance_info(instance)
        except Exception as e:
            self.debug(1,e)
            # we likely have an Internet connection problem ...
            # let's just ignore the situation and continue
            self.debug(1,"INTERNET connection error. The process will stop.")
            return None 

        # this is an Internet error
        if instance.get_state() == CloudSendInstanceState.RUNNING:
            ssh_conn = await self._connect_to_instance(instance)
            if ssh_conn is None:
                self.debug(1,"FATAL ERROR(0):",msgs,instance,color=bcolors.FAIL)
                return None 
            self.debug(2,"HANDLE_DISCONNECT: LIKELY INTERNET ERROR?")
            return ssh_conn

        self.debug(1,"ERROR:",msg,instance,color=bcolors.FAIL)
        if run_session:
            if instance.get_state() & (CloudSendInstanceState.STOPPING | CloudSendInstanceState.STOPPED) :
                # mark any type of process as aborted, but DONE
                self._mark_aborted(run_session,instance,CloudSendProcessState.ANY - CloudSendProcessState.DONE) 
            elif instance.get_state() & (CloudSendInstanceState.TERMINATING | CloudSendInstanceState.TERMINATED):
                # mark any type of process as aborted
                self._mark_aborted(run_session,instance,CloudSendProcessState.ANY) 
            if do_revive:
                await self._revive(run_session,instance)
        ssh_conn = await self._connect_to_instance(instance)
        if ssh_conn is None:
            self.debug(1,"FATAL ERROR(1):",msgs,instance,color=bcolors.FAIL)

        # free the lock
        if do_revive:
            self._instances_reviving[instance.get_name()] = False 

        return ssh_conn 

    def _compute_jobs_info(self,processes):
        jobsinfo = ""

        for process in processes:
            job         = process.get_job()    # deployed job
            dpl_env     = job.get_env()        # deployed job has a deployed environment
            shash       = job.get_hash()
            uid         = process.get_uid()
            pid         = process.get_pid()
            if jobsinfo:
                jobsinfo = jobsinfo + " " + dpl_env.get_name_with_hash() + " " + str(shash) + " " + str(uid) + " " + str(pid) + " \"" + str(job.get_config('output_file')) + "\""
            else:
                jobsinfo = dpl_env.get_name_with_hash() + " " + str(shash) + " " + str(uid) + " " + str(pid) + " \"" + str(job.get_config('output_file')) + "\""
            
        return jobsinfo 
    
    async def __fetch_states_internal( self, run_session, instance , processes , do_revive , ssh_conn ):

        fetched     = dict()

        # nothing to fetch
        if not processes or len(processes)==0:
            return fetched , ssh_conn
    
        global_path = instance.get_global_dir()
        state_sh    = instance.path_join( global_path , 'state.sh' )
        jobsinfo    = self._compute_jobs_info(processes)
        cmd         =  state_sh + " " + jobsinfo
        
        self.debug(2,"PROCESSES sent for state.sh",processes)
        self.debug(2,"Executing command",cmd)
        
        attempts = 0 
        while True:
            try:
                stdout, stderr = await self._exec_command(ssh_conn,cmd)
                data = await stdout.read()
                break
            except (OSError, asyncssh.Error):
                self.debug(1,"SSH connection error while sending state.sh command")
                ssh_conn = await self._handle_instance_disconnect(run_session,instance,do_revive,"could not get jobs states for instance. SSH connection lost with")
                if ssh_conn is None:
                    return None
                # this is part of a run session we're looking at...
                # lets update the processes and subsequent variables after the error
                if run_session:
                    processes = run_session.get_active_processes(instance)
                    jobsinfo  = self._compute_jobs_info(processes)
                    cmd       =  state_sh + " " + jobsinfo

                attempts += 1
            if attempts > 5:
                self.debug(1,"TOO MANY ATTEMPTS",color=bcolors.FAIL)
                return None                

        for line in data.splitlines():      
            if not line or line.strip()=="":
                break     
            statestr = line.strip() 
            self.debug(2,"State=",statestr,"IP=",instance.get_ip_addr())
            stateinfo = statestr.split(',')
            statestr  = re.sub(r'\([0-9]+\)','',stateinfo[2])
            uid       = stateinfo[0]
            pid       = stateinfo[1]

            process = None
            for p in processes:
                if p.get_uid() == uid:
                    process = p
                    break
            if process is not None:
                fetched[uid] = True
                # we don't have PIDs with batches
                # let's take the opportunity to update it here...
                if process.get_pid() is None and pid != "None":
                    process.set_pid( int(pid) )
                try:
                    state = CloudSendProcessState[statestr.upper()]
                    # let's keep as much history as we know 
                    # ABORTED state has more info than UNKNOWN ...
                    # on a new state recovery, the newly created instance is returining UNKNOWN
                    # but if the maestro witnessed an ABORTED state
                    # we prefer to keep this ...
                    if not (process.get_state() == CloudSendProcessState.ABORTED and state == CloudSendProcessState.UNKNOWN):
                        process.set_state(state)
                    self.debug(2,process)
                    # SUPER IMPORTANT TO TEST with retrieved (remote) state !
                    # if we tested with curernt state we could test against all ABORTED jobs and the wait() function would cancel...
                    # this could potentially happen though
                    # instead, the retrieved state is actually UNKNOWN
                    self.debug(2,'The state is ',uid,state)
                except Exception as e:
                    debug(1,"\nUnhandled state received by state.sh!!!",statestr,"\n")
                    debug(2,e)
                    state = CloudSendProcessState.UNKNOWN
            else:
                debug(2,"Received UID info that was not requested")

        return fetched , ssh_conn

    async def __wait_for_state_internal( self , run_session , instance , job_state , wait_mode , daemon ):

        instance_name = instance.get_name()
        try:

            if wait_mode & CloudSendProviderStateWaitMode.WATCH:
                self._instances_watching[instance_name] = True
                self._instances_reviving[instance_name] = False
            
            ssh_conn = await self._connect_to_instance(instance)

            do_revive = (wait_mode & CloudSendProviderStateWaitMode.WATCH)!=0

            if ssh_conn is None:
                ssh_conn = await self._handle_instance_disconnect(run_session,instance,do_revive,"could not get jobs states for instance")
                if ssh_conn is None:
                    return None

            global_path = instance.get_global_dir() 

            while True:

                # update before
                processes = run_session.get_active_processes(instance)

                fetched , ssh_conn = await self.__fetch_states_internal(run_session,instance,processes,do_revive,ssh_conn)

                # always update the activate processes in case something happened
                processes = run_session.get_active_processes(instance)

                # print job status summary
                if not daemon:
                    self.print_jobs_summary(run_session,instance)

                # all retrived attributes need to be true
                arr_retrieved = [ fetched.get(process.get_uid(),False) for process in processes]
                arr_test      = [ (process.get_state() & job_state)!=0 for process in processes]
                retrieved = all( arr_retrieved )
                tested    = all( arr_test )

                self.debug(2,retrieved,arr_retrieved)
                self.debug(2,tested   ,arr_test     )

                if retrieved and tested :
                    break

                self.serialize_state()

                if wait_mode & CloudSendProviderStateWaitMode.WAIT:
                    try:
                        await asyncio.sleep(SLEEP_PERIOD)
                    except asyncio.exceptions.CancelledError:
                        self.debug(1,"Wait process cancelled",color=bcolors.WARNING)
                else:
                    break

            ssh_conn.close() 

            self.serialize_state()

            if wait_mode & CloudSendProviderStateWaitMode.WATCH:
                # lets wait 1 minutes before stopping
                # this helps with the demo which runs a wait() and a get() sequentially ...
                if self._auto_stop:
                    await asyncio.sleep(60*1)  

                self._instances_watching[instance.get_name()] = False            
                any_watching = any( self._instances_watching.values() )

                if not any_watching:
                    self.debug(2,"WATCHING has ended")
                    run_session.deactivate()
                    # we mark it specifically as IDLE so we know it's been ran once ... (could be useful)
                    self.set_state(  (CloudSendProviderState.IDLE | self._state) & (CloudSendProviderState.ANY - CloudSendProviderState.WATCHING - CloudSendProviderState.RUNNING) )
                    self.debug(2,"entering IDLE state",self._state)

                if self._auto_stop:
                    try:
                        self.debug(1,"Stopping instance",instance.get_name(),color=bcolors.WARNING)
                        self.stop_instance(instance)
                    except:
                        pass
                    debug(2,self._instances_watching)
                    if not any_watching:
                        self.debug(1,"Stopping the fat client (maestro) because all instances have ran the jobs",color=bcolors.WARNING)
                        os.system("sudo shutdown -h now")

        except Exception as e:
            # make sure we catch any exception and unlock those variables
            if wait_mode & CloudSendProviderStateWaitMode.WATCH:
                self._instances_watching[instance_name] = False
                self._instances_reviving[instance_name] = False
            raise e

    def _get_num_workers(self):
        num_workers = 10
        if self._instances:
            num_workers = len(self._instances)
        return num_workers

    async def __wait_jobs_state( self, run_session , wait_state = CloudSendProviderStateWaitMode.NO_WAIT , job_state = CloudSendProcessState.ANY , daemon = False ):   

        if not run_session:
            return None

        jobs = []
        for instance in self._instances:
            jobs.append( self.__wait_for_state_internal(run_session,instance,job_state,wait_state,daemon) )

        if not daemon:
            await asyncio.gather( *jobs )
        else:
            await self._cancel_watch()
            # spawn it ...
            self._watcher_task = asyncio.ensure_future( asyncio.gather( *jobs ) )            
            self.debug(2,self._watcher_task)
            if wait_state & CloudSendProviderStateWaitMode.WATCH:
                self.debug(2,"Watching ...")
            await asyncio.sleep(2) # let it go to the loop
        # done

    async def wakeup(self):
        # if self._state != CloudSendProviderState.WATCHING:
        #     self.debug(1,"Provider was not watching: cancelling automatic wakeup")
        #     return 
        # else:
        if self._recovery:
            if self._state & CloudSendProviderState.STARTED:
                await self.start()
            if self._state & CloudSendProviderState.ASSIGNED:
                await self._assign()
            if self._state & CloudSendProviderState.DEPLOYED:
                await self.deploy()
            # should we have those here ? Or just let watch do it's thing ? 
            # actually, thanks to state recovery, run_jobs should be smart enough to not run DONE jobs again...
            if self._state & CloudSendProviderState.RUNNING:
                await self.run()
        else:
            # self.start()
            # self._assign()
            # self.deploy()
            pass

    async def _cancel_watch(self):
        if self._watcher_task is not None:
            self._watcher_task.cancel()
            try:
                await self._watcher_task
                self._watcher_task = None
            except asyncio.CancelledError:
                pass


    def is_watching(self):
        res = self._watcher_task is not None 
        res = res and not self._watcher_task.cancelled() and not self._watcher_task.done() 
        res = res and self._state & CloudSendProviderState.WATCHING 
        return res

    async def _watch(self):

        if not self._state & CloudSendProviderState.STARTED:
            self.debug(1,"Not ready to watch. Call 'start' first",color=bcolors.WARNING)
            return

        job_state = CloudSendProcessState.DONE|CloudSendProcessState.ABORTED
        
        # switch the state to watch mode ... 
        # this will allow to check if the Provider needs to run all methods until watch, on wakeup
        # (no matter the state recovery)
        self.set_state( self._state | CloudSendProviderState.WATCHING )

        await self.__wait_jobs_state(self._current_session , CloudSendProviderStateWaitMode.WAIT|CloudSendProviderStateWaitMode.WATCH,job_state,True)

    async def _wait(self,job_state,run_session,instance):
        while True:

            # with this new in-memory test, we can't run the test while the instance is reviving
            # because the activate processes are being changed ...
            while self._instances_reviving.get(instance.get_name(),False):
                self.debug(1,"waiting for reviving instance (1)",instance)
                await asyncio.sleep(SLEEP_PERIOD)

            self.print_jobs_summary(run_session,instance)

            if not run_session:
                break
            
            test = True 
            for process in run_session.get_active_processes(instance):
                test = test and (process.get_state() & job_state )
            
            if test:
                break
            
            await asyncio.sleep(SLEEP_PERIOD)

    
    async def wait(self,job_state,run_session=None):
        
        await self._ensure_watch()

        # let the watcher fetch the first time states ...
        await asyncio.sleep(1)

        if run_session is None:
            run_session = self._current_session

        # OLD METHOD
        # we used to fetch the same way watch was fetching but this is redundant
        # and it was causing issues between lists of processes etc.
        #return await self.__get_or_wait_jobs_state(CloudSendProviderStateWaitMode.WAIT,job_state)
        
        # NEW METHOD
        jobs = []
        for instance in self._instances:
            jobs.append( self._wait(job_state,run_session,instance) )
        await asyncio.gather( *jobs )

    async def __get_jobs_states_internal(self,run_session,instance):
        instanceid , ssh_conn , ftp_client = await self._wait_and_connect(instance)

        # no recovery here , this is an observation call ... we just fail ...
        if ssh_conn is None:
            self.debug(1,"ERROR: could not get jobs states",instance,color=bcolors.FAIL)
            return

        # let's update the active processes
        processes = run_session.get_active_processes(instance)
        # first argument: run_session = None, so this won't trigger re-runs etc....
        await self.__fetch_states_internal(None,instance,processes,False,ssh_conn)

        # print
        self.print_jobs_summary(run_session,instance)

        ssh_conn.close()


    async def get_jobs_states(self,run_session=None):

        self.debug(2,'Getting Jobs States ...')

        if not self._state & CloudSendProviderState.STARTED:
            self.debug(1,"Not ready to get job states. Call 'start' first")
            return

        if not run_session:
            run_session = self._current_session

        jobs = []
        for instance in self._instances:
            jobs.append( self.__get_jobs_states_internal(run_session,instance) )
        
        await asyncio.gather( *jobs )


    @abstractmethod
    def get_recommended_cpus(self,inst_cfg):
        pass

    @abstractmethod
    def get_cpus_cores(self,inst_cfg):
        pass


from enum import IntFlag
from abc import ABC , abstractmethod
import cloudsend.utils as cloudsendutils
import json
import copy
import os

cs_keypairName         = 'cloudsend-keypair'
cs_secGroupName        = 'cloudsend-sec-group-allow-ssh'
cs_secGroupNameMaestro = 'cloudsend-sec-group-allow-maestro'
cs_bucketName          = 'cloudsend-bucket'
cs_vpcName             = 'cloudsend-vpc'
cs_instanceNameRoot    = 'cloudsend-instance'
cs_instanceMaestro     = 'cloudsend-maestro'
cs_environmentNameRoot = 'cloudsend-env'
cs_maestroProfileName  = 'cloudsend-maestro-profile'
cs_maestroRoleName     = 'cloudsend-maestro-role'
cs_maestroPolicyName   = 'cloudsend-maestro-policy'

K_LOADED   = '_loaded_'
K_COMPUTED = '_computed_'

# NEW > STARTED > ASSIGNED > DEPLOYED > ( RUNNING | WATCHING <-> IDLE )

class CloudSendProviderState(IntFlag):
    NEW           = 0  # provider created
    STARTED       = 1  # provider started
    ASSIGNED      = 2  # provider assigned jobs
    DEPLOYED      = 4  # provider deployed  
    RUNNING       = 8  # provider ran jobs
    WATCHING      = 16 # provider is watching jobs
    IDLE          = 32 # provider has stopped running jobs and watching
    ANY           = 32 + 16 + 8 + 4 + 2 + 1     

class bcolors:
    HEADER = '\033[95m'
    OKBLUE = '\033[94m'
    OKCYAN = '\033[96m'
    OKGREEN = '\033[92m'
    WARNING = '\033[93m'
    FAIL = '\033[91m'
    ENDC = '\033[0m'
    BOLD = '\033[1m'
    UNDERLINE = '\033[4m'

    CBLACK  = '\33[30m'
    CRED    = '\33[31m'
    CGREEN  = '\33[32m'
    CYELLOW = '\33[33m'
    CBLUE   = '\33[34m'
    CVIOLET = '\33[35m'
    CBEIGE  = '\33[36m'
    CWHITE  = '\33[37m'

    CBLACKBG  = '\33[40m'
    CREDBG    = '\33[41m'
    CGREENBG  = '\33[42m'
    CYELLOWBG = '\33[43m'
    CBLUEBG   = '\33[44m'
    CVIOLETBG = '\33[45m'
    CBEIGEBG  = '\33[46m'
    CWHITEBG  = '\33[47m'        

class CloudSendError(Exception):
    pass

class CloudSendInstanceState(IntFlag):
    UNKNOWN     = 0
    STARTING    = 1  # instance starting
    RUNNING     = 2  # instance running
    STOPPING    = 4  # stopping
    STOPPED     = 8  # stopped
    TERMINATING = 16 # terminating
    TERMINATED  = 32 # terminated
    ANY         = 32 + 16 + 8 + 4 + 2 + 1     


class CloudSendProcessState(IntFlag):
#    FOO = 100
    UNKNOWN   = 0
    WAIT      = 1  # waiting for bootstraping
    QUEUE     = 2  # queued (for sequential scripts)
    IDLE      = 4  # script about to start
    RUNNING   = 8  # script running
    DONE      = 16 # script has completed
    ABORTED   = 32 # script has been aborted
    ANY       = 32 + 16 + 8 + 4 + 2 + 1 

class CloudSendPlatform(IntFlag):
    LINUX       = 1
    WINDOWS     = 2
    WINDOWS_WSL = 3

class CloudSendInstance():

    def __init__(self,config,id,proprietaryData=None):
        # instance region
        self._region   = config.get('region')
        # naming
        self._name     = init_instance_name(config)
        self._id       = id
        self._rank     = config.get('rank',"1.1")
        # IP / DNS
        self._ip_addr  = None
        self._dns_addr = None
        self._ip_addr_priv = None
        self._dns_addr_priv = None
        # state
        self._state    = CloudSendProcessState.UNKNOWN
        # the config the instance has been created on
        self._config   = config 
        # dict data associated with it (AWS response data e.g.)
        self._data     = proprietaryData
        # jobs list
        self._jobs     = [ ]
        # env dict
        self._envs     = dict()
        # invalid
        self._invalid  = False
        self._platform = CloudSendPlatform.LINUX
        self._reachability = False

    def get_region(self):
        return self._region

    def get_id(self):
        return self._id 
     
    def get_name(self):
        return self._name

    def get_rank(self):
        return self._rank

    def get_ip_addr(self):
        return self._ip_addr

    def get_dns_addr(self):
        return self._dns_addr 

    def get_ip_addr_priv(self):
        return self._ip_addr_priv

    def get_dns_addr_priv(self):
        return self._dns_addr_priv

    def get_reachability(self):
        return self._reachability

    def get_cpus(self):
        return self._config.get('cpus')

    def get_state(self):
        return self._state

    def get_platform(self):
        return self._platform

    def get_active_processes(self):
        processes_res = []
        for job in self._jobs:
            dpl_jobs = job.get_deployed_jobs()
            for dpl_job in dpl_jobs:
                for process in dpl_job.get_processes():
                    if process.is_active():
                        processes_res.append(process)
        return processes_res

    def get_home_dir(self,absolute=True):
        if self._platform == CloudSendPlatform.LINUX or self._platform == CloudSendPlatform.WINDOWS_WSL:
            return '/home/' + self.get_config('img_username') if absolute else '%HOME'
        elif self._platform == CloudSendPlatform.WINDOWS:
            return 'C:\>' + self.get_config('img_username') if absolute else '%HOME%'

    def get_global_dir(self):
        return self.path_join( self.get_home_dir() , 'run' )

    def path_join(self,*args):
        if self._platform == CloudSendPlatform.LINUX or self._platform == CloudSendPlatform.WINDOWS_WSL:
            return '/'.join(args)
        elif self._platform == CloudSendPlatform.WINDOWS:
            return '\\'.join(args)

    def path_dirname(self,path):
        #return os.path.dirname(path)
        sep = self.path_sep()
        path_expl = path.split(sep)
        if len(path_expl)>2:
            return sep.join( path_expl[:-1] )
        else:
            return ''
    
    def path_basename(self,path):
        #return os.path.basename(path)
        sep = self.path_sep()
        path_expl = path.split(sep)
        if len(path_expl)>1:
            return path_expl[-1]
        else:
            return ''


    def path_sep(self):
        if self._platform == CloudSendPlatform.LINUX or self._platform == CloudSendPlatform.WINDWS_WSL:
            return '/'
        elif self._platform == CloudSendPlatform.WINDWS:
            return '\\'

    def set_ip_addr(self,value):
        self._ip_addr = value

    def set_ip_addr_priv(self,value):
        self._ip_addr_priv = value

    def set_dns_addr(self,value):
        self._dns_addr = value
     
    def set_dns_addr_priv(self,value):
        self._dns_addr_priv = value

    def set_state(self,value):
        self._state = value 

    def set_reachability(self,value):
        self._reachability = value

    def set_platform(self,value):
        self._platform = value 

    def set_invalid(self,val):
        self._invalid = val

    def set_data(self,data):
        self._data = data 
    
    def get_data(self,key):
        if not self._data:
            return None
        return self._data.get(key,None)
     
    def get_config(self,key):
        if not self._config:
            return None
        return self._config.get(key,None)

    def reset_jobs(self):
        self._jobs = []

    def append_job(self,job):
        self._jobs.append(job)
        env = job.get_env()
        self._envs[env.get_name()] = env 

    def get_environments(self):
        return self._envs.values()

    def get_jobs(self):
        return self._jobs

    def get_active_processes(self):
        processes_res = []
        for job in self._jobs:
            for p in job.get_active_processes():
                processes_res.append(p)
        return processes_res

    def get_config_DIRTY(self):
        return self._config

    def is_invalid(self):
        return self._invalid

    def update_from_instance(self,instance):
        self._region   = instance._region
        self._name     = instance._name
        self._id       = instance._id 
        self._rank     = instance._rank
        self._ip_addr  = instance._ip_addr
        self._dns_addr = instance._dns_addr
        self._ip_addr_priv  = instance._ip_addr_priv
        self._dns_addr_priv = instance._dns_addr_priv
        self._state    = instance._state
        self._reachability  = instance._reachability
        self._platform = instance._platform
        self._config   = copy.deepcopy(instance._config)
        self._data     = copy.deepcopy(instance._data)        

    def __repr__(self):
        # return "{0}: REGION = {1} , ID = {2} , NAME = {3} , IP = {4} , CPUS = {5} , RANK = {6}".format(type(self).__name__,self._region,self._id,self._name,self._ip_addr,self.get_cpus(),self._rank)
        # return "{0}: ID = {1} , NAME = {2} , IP = {3} , CPUS = {4}".format(type(self).__name__,self._id,self._name,self._ip_addr,self.get_cpus())
        return "{0}: NAME = {1} , IP = {2}".format(type(self).__name__,self._name,self._ip_addr)

    def __str__(self):
        # return "{0}: REGION = {1} , ID = {2} , NAME = {3} , IP = {4} , CPUS = {5} , RANK = {6}".format(type(self).__name__,self._region,self._id,self._name,self._ip_addr,self.get_cpus(),self._rank)
        # return "{0}: ID = {1} , NAME = {2} , IP = {3} , CPUS = {4}".format(type(self).__name__,self._id,self._name,self._ip_addr,self.get_cpus())
        return "{0}: NAME = {1} , IP = {2}".format(type(self).__name__,self._name,self._ip_addr)


class CloudSendEnvironment():

    def __init__(self,projectName,env_config):
        self._config   = env_config
        self._project  = projectName
        _env_obj       = cloudsendutils.compute_environment_object(env_config)
        self._hash     = cloudsendutils.compute_environment_hash(_env_obj)
        if not self._config.get('name'):
            self._name = cs_environmentNameRoot

            append_str = '-' + self._hash
            if env_config.get('dev') == True:
                append_str = ''
            if projectName:
                self._name = cs_environmentNameRoot + '-' + projectName + append_str
            else:
                self._name = cs_environmentNameRoot + append_str
        else:
            self._name = self._config.get('name')

    def get_name(self):
        return self._name

    def get_name_with_hash(self):
        return self._name + '-' + self._hash

    def get_config(self,key):
        return self._config.get(key)

    def get_env_obj(self):
        _env_obj = cloudsendutils.compute_environment_object(self._config)
        return _env_obj 

    def json(self):
        return json.dumps(self.get_env_obj())          

    def deploy(self,instance):
        return CloudSendDeployedEnvironment(self,instance)

    def __repr__(self):
        return "{0}: NAME = {1} , HASH = {2}".format(type(self).__name__,self._name,self._hash)

    def __str__(self):
        return "{0}: NAME = {1} , HASH = {2}".format(type(self).__name__,self._name,self._hash)


# "Temporary" objects used when starting scripts      

class CloudSendDeployedEnvironment(CloudSendEnvironment):

    # constructor by copy...
    def __init__(self, env, instance):
        #super().__init__( env._project , env._config )
        self._config   = env._config #copy.deepcopy(env._config)
        self._project  = env._project
        self._hash     = env._hash
        self._name     = env._name
        self._instance = instance
        self._path     = instance.path_join( instance.get_home_dir() , 'run' , self.get_name_with_hash())

    def get_path(self):
        return self._path

    def get_instance(self):
        return self._instance

    def json(self):
        _env_obj = super().get_env_obj()
        # overwrite name in conda config as well
        if _env_obj['env_conda'] is not None:
           _env_obj['env_conda']['name'] = self.get_name_with_hash()
        _env_obj['name'] = self.get_name_with_hash()
        # replace __REQUIREMENTS_TXT_LINK__ with the actual requirements.txt path (dependent of config and env hash)
        # the file needs to be absolute
        requirements_txt_path = self._instance.path_join(self._path,'requirements.txt')
        _env_obj = cloudsendutils.update_requirements_path(_env_obj,requirements_txt_path)
        return json.dumps(_env_obj)  


class CloudSendJob():

    def __init__(self,job_cfg,rank):
        self._config    = job_cfg
        self._rank      = rank
        self._hash      = cloudsendutils.compute_job_hash(self._config)
        self._env       = None
        self._instance  = None
        self._deployed = [ ]
        if (not 'input_file' in self._config) or (not 'output_file' in self._config) or not isinstance(self._config['input_file'],str) or not isinstance(self._config['output_file'],str):
            print("\n\n\033[91mConfiguration requires an input and output file names\033[0m\n\n")
            raise CloudSendError() 

    def attach_env(self,env):
        self._env = env 
        self._config['env_name'] = env.get_name()

    def get_hash(self):
        return self._hash

    def get_config(self,key,defaultVal=None):
        return self._config.get(key,defaultVal)

    def get_env(self):
        return self._env

    def get_instance(self):
        return self._instance

    def get_rank(self):
        return self._rank

    def get_deployed_jobs(self):
        return self._deployed

    def get_deployed_job(self,instance):
        for dpl_job in self._deployed:
            if dpl_job.get_instance() == instance:
                return dpl_job
        return None

    def get_active_processes(self):
        processes_res = []
        for dpl_job in self._deployed:
            for p in dpl_job.get_active_processes():
                processes_res.append(p)
        return processes_res

    def deploy(self,dpl_env,add_permanently=True):
        # instance = dpl_env.get_instance()
        # dpl_job  = self.get_deployed_job(instance)
        # if dpl_job:
        #     return dpl_job
        dpl_job  = CloudSendDeployedJob(self,dpl_env)
        if add_permanently:
            self._deployed.append(dpl_job)
        return dpl_job

    def has_completed(self):
        for dpl_job in self._deployed:
            for process in dpl_job.get_processes():
                if process.get_state() == CloudSendProcessState.DONE:
                    return True
        return False

    def get_last_process(self):
        if not self._deployed or len(self._deployed)==0:
            return None
        # last_dpl_job = self._deployed[len(self._deployed)-1]
        # processes = last_dpl_job.get_processes()
        # if not processes or len(processes)==0:
        #     return None 
        # return processes[len(processes)-1]
        last_process = None
        count = 0 
        for dpl_job in self._deployed:
            for process in dpl_job.get_processes():
                last_process = process
                count = count + 1
        return last_process

    def set_instance(self,instance):
        self._instance = instance
        instance.append_job(self)

    def str_simple(self):
        return "{0}: HASH = {1} , ENV = {2}".format(type(self).__name__,self.get_hash(),self.get_env().get_name() if self.get_env() else None)

    def __repr__(self):
        return "{0}: HASH = {1} , INSTANCE = {2} , ENV = {3}".format(type(self).__name__,self.get_hash(),self.get_instance(),self.get_env().get_name() if self.get_env() else None)
         
    def __str__(self):
        return "{0}: HASH = {1} , INSTANCE = {2} , ENV = {3}".format(type(self).__name__,self.get_hash(),self.get_instance(),self.get_env().get_name() if self.get_env() else None)


# "Temporary" objects used when starting scripts     
# "Proxy" class that keeps the link with "copied" object
# We proxy all parent methods instead of using inheritance
# this allows to keep the same behavior while keeping the link and sharing memory objects

class CloudSendDeployedJob(CloudSendJob):

    def __init__(self,job,dpl_env):
        #super().__init__( job._config )
        self._job       = job
        #self._config    = copy.deepcopy(job._config)
        #self._hash      = job._hash
        #self._instance  = job._instance
        self._processes = []
        self._env       = dpl_env
        self._instance  = dpl_env.get_instance()
        self._path      = self._instance.path_join( dpl_env.get_path() , self.get_hash() )
        self._command   = cloudsendutils.compute_job_command(self._instance,self._path,self._job._config)

    def attach_process(self,process):
        self._processes.append(process)

    def get_path(self):
        return self._path

    def get_command(self):
        return self._command

    def attach_env(self,env):
        raise CloudSendError('Can not attach env to deployed job')

    # proxied
    def get_rank(self):
        return self._job._rank

    # proxied
    def get_hash(self):
        return self._job._hash

    # proxied
    def get_config(self,key,defaultVal=None):
        return self._job._config.get(key,defaultVal)

    def get_env(self):
        return self._env

    def get_instance(self):
        return self._instance   

    def get_processes(self):
        return self._processes

    def get_active_processes(self):
        processes_res = []
        for p in self._processes:
            if p.is_active():
                processes_res.append(p)
        return processes_res

    def deploy(self,dpl_env):
        raise CloudSendError('Can not deploy a deployed job')

    def set_instance(self,instance):
        raise CloudSendError('Can not set the instance of a deployed job')


class CloudSendProcess():

    def __init__(self,dpl_job,batch,pid=None):
        self._job   = dpl_job
        self._uid   = cloudsendutils.generate_unique_id() 
        self._pid   = pid
        self._batch = batch
        self._state = CloudSendProcessState.UNKNOWN
        self._active = True
        self._job.attach_process(self)
     
    def get_uid(self):
        return self._uid

    def get_pid(self):
        return self._pid

    def get_batch(self):
        return self._batch 

    def deactivate(self):
        self._active = False

    def is_active(self):
        return self._active 

    def get_path(self):
        instance = self._job.get_instance()
        return instance.path_join( self._job.get_path() , self._uid )

    def get_state(self):
        return self._state
     
    def set_state(self,value):
        self._state = value

    def set_pid(self,value):
        self._pid = value 

    def get_job(self):
        return self._job 

    def str_simple(self):
        if self._batch:
            return "CloudSendProcess: UID = {0} , PID = {1} , BATCH = {2} , STATE = {3}".format(self._uid,str(self._pid).rjust(5),self._batch.get_uid(),self._state.name)
        else:
            return "CloudSendProcess: UID = {0} , PID = {1} , STATE = {2}".format(self._uid,str(self._pid).rjust(5),self._state.name)

    def __repr__(self):
        return "CloudSendProcess: job = {0} , UID = {1} , PID = {2} , STATE = {3}".format(self._job,self._uid,str(self._pid).rjust(5),self._state.name)
         
    def __str__(self):
        return "CloudSendProcess: job = {0} , UID = {1} , PID = {2} , STATE = {3}".format(self._job,self._uid,str(self._pid).rjust(5),self._state.name)


class CloudSendRunSession():

    def __init__(self,number):
        self._number  = number 
        self._batches = []

    def create_batch(self):
        batch = CloudSendBatch(self)
        self._batches.append(batch)
        return batch

    def get_active_processes(self,instance=None):
        processes_res = []
        for batch in self._batches:
            for p in batch.get_active_processes(instance):
                processes_res.append(p)
        return processes_res

    def get_number(self):
        return self._number

    # return the processes that have been run for each job
    # (kinda last_processes but through a different view point)
    def get_ran_processes(self):
        processes = dict()
        # scan all the processes in order of the batches ...
        for batch in self._batches:
            for process in batch.get_processes():
                dpl_job = process.get_job()
                rank    = dpl_job.get_rank()
                processes[rank] = process # this will keep the last process ran for this job rank
        
        return processes.values()


    def deactivate(self,instance=None):
        for batch in self._batches:
            batch.deactivate_processes(instance)

    def mark_aborted(self,instance,state_mask):
        for batch in self._batches:
            batch.mark_aborted(instance,state_mask)


class CloudSendBatch():

    def __init__(self,run_session):
        self._uid      = cloudsendutils.generate_unique_id()
        self._session  = run_session
        self._instances_processes = dict()

    def get_uid(self):
        return self._uid

    def get_session(self):
        return self._session

    def create_process(self,dpl_job):
        process   = CloudSendProcess( dpl_job , self )
        instance  = dpl_job.get_instance()
        inst_name = instance.get_name()
        if not inst_name in self._instances_processes:
            self._instances_processes[inst_name] = []
        self._instances_processes[inst_name].append(process)
        return process

    def get_processes(self):
        processes_res = []
        for instance_name , processes in self._instances_processes.items():
            for process in processes:
                processes_res.append(process)
        return processes_res

    def get_active_processes(self,instance=None):
        processes_res = []
        if instance:
            if instance.get_name() in self._instances_processes:
                for process in self._instances_processes[instance.get_name()]:
                    if process.is_active():
                        processes_res.append(process)
        else:
            for instance_name , processes in self._instances_processes.items():
                for process in processes:
                    if process.is_active():
                        processes_res.append(process)
        return processes_res

    def deactivate_processes(self,instance=None):
        for instance_name , processes in self._instances_processes.items():
            if not instance or instance.get_name() == instance_name:
                for process in processes:
                    process.deactivate()

    # mark the currently active process as ABORTED
    def mark_aborted(self,instance,state_mask):
        for process in self.get_active_processes(instance):
            if process.get_state() & state_mask:
                process.set_state(CloudSendProcessState.ABORTED)

def init_instance_name(instance_config):
    
    if 'maestro' not in instance_config:

        if instance_config.get('dev',False)==True:
            append_str = '' 
        else:
            append_str = '-' + cloudsendutils.compute_instance_hash(instance_config)

        if 'rank' not in instance_config:
            debug(1,"\033[93mDeveloper: you need to set dynamically a 'rank' attribute in the config for the new instance\033[0m")
            sys.exit(300) # this is a developer error, this should never happen so we can use exit here
            
        if 'project' in instance_config:
            return cs_instanceNameRoot + '-' + instance_config['project'] + '-' + instance_config['rank'] + append_str
        else:
            return cs_instanceNameRoot + '-' + instance_config['rank'] + append_str
    
    else:

        # if 'project' in instance_config:
        #     return cs_instanceMaestro + '-' + instance_config['project'] 
        # else:
        #     return cs_instanceMaestro

        # ultimately, the maestro will be shared across projects... (to save $)
        return cs_instanceMaestro



# -*- coding: utf-8 -*-
from setuptools import setup

packages = \
['cloudsend', 'cloudsend.resources.remote_files']

package_data = \
{'': ['*'],
 'cloudsend': ['resources/instancetypes-aws.csv',
               'resources/instancetypes-aws.csv',
               'resources/requirements.txt',
               'resources/requirements.txt']}

install_requires = \
['asyncssh>=2.12.0,<2.13.0',
 'boto3>=1,<2',
 'jcs>=0.2.1,<0.3.0',
 'ortools>=9,<10',
 'psutil>=5.9.4,<6.0.0',
 'pyyaml>=6.0,<7.0']

entry_points = \
{'console_scripts': ['cli = cloudsend.cli:main',
                     'client = cloudsend.maestroclient:main',
                     'demo = cloudsend.demo:main']}

setup_kwargs = {
    'name': 'cloudsend',
    'version': '0.4.2',
    'description': 'CloudSend is a Python package that allows you to run any script on a cloud service (for now AWS only).',
    'long_description': '# Description\n\nCloudSend is a Python package that allows you to run any script on a cloud service (for now AWS only).\n\n# Features\n\n- Easily run scripts on AWS by writing a simple configuration file\n- Handles Python and Julia scripts, or any command\n- Handles PyPi , Conda/Mamba, Apt-get and Julia environments\n- Concurrent instance support\n- Handles disconnections from instances, including stopped or terminated instances\n- Handles interruption of CloudSend, with state recovery\n- Runs locally or on a remote instance, with \'watcher\' functionality \n\n# Pre-requisites\n\nIn order to use the python AWS client (Boto3), you need to have an existing AWS account and to setup your computer for AWS.\n\n## with AWS CLI\n\n1. Go to [the AWS Signup page](https://portal.aws.amazon.com/billing/signup#/start/email) and create an account\n2. Download [the AWS CLI](https://docs.aws.amazon.com/cli/latest/userguide/getting-started-install.html)\n3. In the AWS web console, [create a user with administrator privilege](https://docs.aws.amazon.com/streams/latest/dev/setting-up.html)\n4. In the AWS web console, under the AMI section, click on the new user and make sure you create an access key under the tab "Security Credentials". Make sure "Console Password" is Enabled as well\n5. In ther Terminal, use the AWS CLI to setup your configuration:\n```\naws configure\n```\nSee [https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-quickstart.html](here)\n\n6. To run in `remote` mode, you also need to [add the following credentials to your user](https://docs.aws.amazon.com/AWSEC2/latest/UserGuide/iam-roles-for-amazon-ec2.html) (maybe):\n- iam:PassRole\n- iam:CreateRole\n- ec2:AssociateIamInstanceProfile\n- ec2:ReplaceIamInstanceProfileAssociation\n\n## manually\n\n1. Go to [the AWS Signup page](https://portal.aws.amazon.com/billing/signup#/start/email) and create an account\n2. In the AWS web console, [create a user with administrator privilege](https://docs.aws.amazon.com/streams/latest/dev/setting-up.html)\n3. In the AWS web console, under the IAM section, click on the new user and make sure you create an access key under the tab "Security Credentials". Make sure "Console Password" is Enabled as well\n4. Add your new user credentials manually, [in the credentials file](https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html)\n\n##### \'~/.aws/config\' example (\'C:\\Users\\USERNAME\\\\.aws\\config\' on Windows)\n\n```\n[default]\nregion = eu-west-3\noutput = json\n```\n\n##### \'~/.aws/credentials\' example (\'C:\\Users\\USERNAME\\\\.aws\\credentials\' on Windows)\n\n```\n[default]\naws_access_key_id = YOUR_ACCESS_KEY_ID\naws_secret_access_key = YOUR_SECRET_ACCESS_KEY\n```\n\n## Setting up a separate user with least permissions (manually) \n\n1. In the AWS web console, in the IAM service, create a group \'cloudsend-users\' with `AmazonEC2FullAccess` and `IAMFullAccess` permissions\n2. In the AWS web console, in the IAM service, create a user USERNAME attached to the \'cloudsend-users\' group:\n### Step 1\n![add user 1](./images/adduser1.jpg)\n### Step 2\n![add user 2](./images/adduser2.jpg)\n### ... Step 5\n![add user 3](./images/adduser3.jpg)\n\n  **COPY the Access Key info !**\n\n3. Add your new user profile manually, [in the credentials file](https://docs.aws.amazon.com/cli/latest/userguide/cli-configure-profiles.html)\n\n##### \'~/.aws/config\' example (\'C:\\Users\\USERNAME\\\\.aws\\config\' on Windows)\n\n```\n[default]\nregion = eu-west-3\noutput = json\n\n[profile cloudsend]\nregion = eu-west-3\noutput = json\n```\n\n##### \'~/.aws/credentials\' example (\'C:\\Users\\USERNAME\\\\.aws\\credentials\' on Windows)\n\n```\n[default]\naws_access_key_id = YOUR_ACCESS_KEY_ID\naws_secret_access_key = YOUR_SECRET_ACCESS_KEY\n\n[cloudsend]\naws_access_key_id = YOU_PROFILE_ACCESS_KEY_ID\naws_secret_access_key = YOUR_PROFILE_SECRET_ACCESS_KEY\n```\n\n4. add the \'profile\' : \'cloudsend_USERNAME\' to the configuration\n\n```python\nconfig = {\n\n    ################################################################################\n    # GLOBALS\n    ################################################################################\n\n    \'project\'      : \'test\' ,                             # this will be concatenated with the instance hashes (if not None) \n    \'profile\'      : \'cloudsend\' ,\n    ...\n```\n\n\n# Installation\n\n## On MacOS / Linux\n\n### with pip\n\n```bash\npython3 -m venv .venv\nsource ./.venv/bin/activate\npython3 -m ensurepip --default-pip\npython3 -m pip install -r requirements.txt\n```\n\n### with poetry\n\n```bash\ncurl -sSL https://install.python-poetry.org | python3.8 -\npoetry install\n```\n\n## On Windows (powershell)\n\n### with pip\n\n```bat\nC:\\> python3 -m venv .venv\nC:\\> .venv\\\\Scripts\\\\activate.bat\nC:\\> python -m pip install -r requirements.txt\n```\n\n### with poetry\n\n```bat\nC:\\> (Invoke-WebRequest -Uri https://install.python-poetry.org -UseBasicParsing).Content | py -\nC:\\> poetry install\n```\n\n# Usage / Test runs\n\n```bash\n# copy the example file\ncp example/config.example.py config.py\n#\n# EDIT THE FILE\n#\n\n# to run with pip\npython3 -m cloudsend.demo config\n# to run with pip with reset (maestro and instances)\npython3 -m cloudsend.demo config reset\n# to run with poetry\npoetry run demo config\n# to run with poetry with reset (maestro and the instances)\npoetry run demo config reset\n```\n\n# Configuration example\n\n```python\nconfig = {\n\n    ################################################################################\n    # GLOBALS\n    ################################################################################\n\n    \'project\'      : \'test\' ,                             # this will be concatenated with the instance hashes (if not None) \n    \'profile\'      : None ,                               # if you want to use a specific profile (user/region), specify its name here\n    \'dev\'          : False ,                              # When True, this will ensure the same instance and dev environement are being used (while working on building up the project) \n    \'debug\'        : 1 ,                                  # debug level (0...3)\n    \'maestro\'      : \'local\' ,                            # where the \'maestro\' resides: local\' | \'remote\' (micro instance)\n    \'auto_stop\'    : True ,                               # will automatically stop the instances and the maestro, once the jobs are done\n    \'provider\'     : \'aws\' ,                              # the provider name (\'aws\' | \'azure\' | ...)\n    \'job_assign\'   : None ,                               # algorithm used for job assignation / task scheduling (\'random\' | \'multi_knapsack\')\n    \'recover\'      : True ,                               # if True, CloudSend will always save the state and try to recover this state on the next execution\n    \'print_deploy\' : False ,                              # if True, this will cause the deploy stage to print more (and lock)\n    \'mutualize_uploads\' : True ,                          # adjusts the directory structure of the uploads ... (False = per job or True = global/mutualized)\n\n\n    ################################################################################\n    # INSTANCES / HARDWARE\n    ################################################################################\n\n    \'instances\' : [\n        { \n            \'region\'       : None ,                       # can be None or has to be valid. Overrides AWS user region configuration.\n            \'cloud_id\'     : None ,                       # can be None, or even wrong/non-existing - then the default one is used\n            \'img_id\'       : \'ami-077fd75cd229c811b\' ,    # OS image: has to be valid and available for the profile (user/region)\n            \'img_username\' : \'ubuntu\' ,                   # the SSH user for the image\n            \'type\'         : \'t2.micro\' ,                 # proprietary size spec (has to be valid)\n            \'cpus\'         : None ,                       # number of CPU cores\n            \'gpu\'          : None ,                       # the proprietary type of the GPU \n            \'disk_size\'    : None ,                       # the disk size of this instance type (in GB)\n            \'disk_type\'    : None ,                       # the proprietary disk type of this instance type: \'standard\', \'io1\', \'io2\', \'st1\', etc\n            \'eco\'          : True ,                       # eco == True >> SPOT e.g.\n            \'eco_life\'     : None ,                       # lifecycle of the machine in ECO mode (datetime.timedelta object) (can be None with eco = True)\n            \'max_bid\'      : None ,                       # max bid ($/hour) (can be None with eco = True)\n            \'number\'       : 1 ,                          # multiplicity: the number of instance(s) to create\n            \'explode\'      : True                         # multiplicity: can this instance type be distributed accross multiple instances, to split CPUs\n        }\n\n    ] ,\n\n    ################################################################################\n    # ENVIRONMENTS / SOFTWARE\n    ################################################################################\n\n    \'environments\' : [\n        {\n            \'name\'         : None ,                       # name of the environment - should be unique if not \'None\'. \'None\' only when len(environments)==1\n\n            # env_conda + env_pypi  : mamba is used to setup the env (pip dependencies included)\n            # env_conda (only)      : mamba is used to setup the env\n            # env_pypi  (only)      : venv + pip is used to setup the env \n\n            \'command\'      : \'example/install_julia.sh\' ,      # None, or a string: path to a bash file to execute when deploying\n            \'env_aptget\'   : [ "openssh-client"] ,        # None, an array of librarires/binaries for apt-get\n            \'env_conda\'    : "example/environment.yml",   # None, an array of libraries, a path to environment.yml  file, or a path to the root of a conda environment\n            \'env_conda_channels\' : None ,                 # None, an array of channels. If None (or absent), defaults and conda-forge will be used\n            \'env_pypi\'     : "example/requirements.txt" , # None, an array of libraries, a path to requirements.txt file, or a path to the root of a venv environment \n            \'env_julia\'    : [ "Wavelets" ] ,             # None, a string or an array of Julia packages to install (requires julia)\n        }\n    ] ,\n\n    ################################################################################\n    # JOBS / SCRIPTS\n    ################################################################################\n\n    \'jobs\' : [\n        {\n            \'env_name\'     : None ,                       # the environment to use (can be \'None\' if solely one environment is provided above)\n            \'cpus_req\'     : None ,                       # the CPU(s) requirements for the process (can be None)\n            \'run_script\'   : \'example/run_remote.py 1 10\',# the script to run (Python (.py) or Julia (.jl) for now) (prioritised vs \'run_command\')\n            \'run_command\'  : None ,                       # the command to run\n            \'upload_files\' : [ "uploaded.txt"] ,          # any file to upload (array or string) - will be put in the same directory\n            \'input_file\'   : \'input.dat\' ,                # the input file name (used by the script)\n            \'output_file\'  : \'output.dat\' ,               # the output file name (used by the script)\n            \'repeat\'       : 2 ,                          # the number of times this job is repeated\n        } ,\n        {\n            \'env_name\'     : None ,                       # the environment to use (can be \'None\' if solely one environment is provided above)\n            \'cpus_req\'     : None ,                       # the CPU(s) requirements for the process (can be None)\n            \'run_script\'   : \'example/run_remote.py 2 12\',# the script to run (Python (.py) or Julia (.jl) for now) (prioritised vs \'run_command\')\n            \'run_command\'  : None ,                       # the command to run\n            \'upload_files\' : [ "uploaded.txt"] ,          # any file to upload (array or string) - will be put in the same directory\n            \'input_file\'   : \'input.dat\' ,                # the input file name (used by the script)\n            \'output_file\'  : \'output.dat\' ,               # the output file name (used by the script)\n        }\n    ]\n}\n```\n\n# Minimum configuration example\n\n```python\nconfig = {\n    \'debug\'        : 1 ,                                  # debug level (0...3)\n    \'maestro\'      : \'local\' ,                            # where the \'maestro\' resides: local\' | \'remote\' (nano instance) | \'lambda\'\n    \'provider\'     : \'aws\' ,                              # the provider name (\'aws\' | \'azure\' | ...)\n\n    \'instances\' : [\n        { \n            \'type\'         : \'t2.micro\' ,                 # proprietary size spec (has to be valid)\n        }\n\n    ] ,\n\n    \'environments\' : [\n        {\n            \'name\'         : None ,                       # name of the environment - should be unique if not \'None\'. \'None\' only when len(environments)==1\n            \'env_conda\'    : "example/environment.yml",   # None, an array of libraries, a path to environment.yml  file, or a path to the root of a conda environment\n            \'env_julia\'    : ["Wavelets"] ,                       # None, a string or an array of Julia packages to install (requires julia)\n        }\n    ] ,\n\n    \'jobs\' : [\n        {\n            \'env_name\'     : None ,                       # the environment to use (can be \'None\' if solely one environment is provided above)\n            \'cpus_req\'     : None ,                       # the CPU(s) requirements for the process (can be None)\n            \'run_script\'   : \'example/run_remote.py 1 10\',# the script to run (Python (.py) or Julia (.jl) for now) (prioritised vs \'run_command\')\n            \'upload_files\' : [ "uploaded.txt"] ,          # any file to upload (array or string) - will be put in the same directory\n            \'input_file\'   : \'input.dat\' ,                # the input file name (used by the script)\n            \'output_file\'  : \'output.dat\' ,               # the output file name (used by the script)\n        } ,\n        {\n            \'env_name\'     : None ,                       # the environment to use (can be \'None\' if solely one environment is provided above)\n            \'cpus_req\'     : None ,                       # the CPU(s) requirements for the process (can be None)\n            \'run_script\'   : \'example/run_remote.py 2 12\',# the script to run (Python (.py) or Julia (.jl) for now) (prioritised vs \'run_command\')\n            \'upload_files\' : [ "uploaded.txt"] ,          # any file to upload (array or string) - will be put in the same directory\n            \'input_file\'   : \'input.dat\' ,                # the input file name (used by the script)\n            \'output_file\'  : \'output.dat\' ,               # the output file name (used by the script)\n        }\n    ]\n}\n```\n\n\n# Python API\n\n```python\nclass CloudSendProcessState(IntFlag):\n    UNKNOWN   = 0\n    WAIT      = 1  # waiting for bootstraping\n    QUEUE     = 2  # queued (for sequential scripts)\n    IDLE      = 4  # script about to start\n    RUNNING   = 8  # script running\n    DONE      = 16 # script has completed\n    ABORTED   = 32 # script has been aborted\n    ANY       = 32 + 16 + 8 + 4 + 2 + 1 \n\nclass CloudSendLightProvider(ABC):\nclass CloudSendFatProvider(ABC):\n\n    def debug(self,level,*args,**kwargs):\n\n    # start the provider: creates the instances\n    # if reset = True, CloudSend forces a process cleanup as well as more re-uploads\n    def start(self,reset):\n\n    # deploy all materials (environments, files, scripts etc.)\n    def deploy(self):\n\n    # run the jobs\n    # returns a CloudSendRunSession\n    def run(self,wait=False):\n\n    # wait for the processes to reach a state\n    def wait(self,job_state,run_session=None):\n\n    # get the states of the processes\n    def get_jobs_states(self,run_session=None):\n\n    def print_jobs_summary(self,run_session=None,instance=None):\n\n    def print_aborted_logs(self,run_session=None,instance=None):\n\n    def fetch_results(self,out_directory,run_session=None):\n\n    # wakeup = start + assign + deploy + run + watch\n    def wakeup(self)\n\n    @abstractmethod\n    def get_region(self):\n\n    @abstractmethod\n    def get_recommended_cpus(self,inst_cfg):\n\n    @abstractmethod\n    def create_instance_objects(self,config,for_maestro):\n\n    @abstractmethod\n    def find_instance(self,config):\n\n    @abstractmethod\n    def start_instance(self,instance):\n\n    @abstractmethod\n    def stop_instance(self,instance):\n\n    @abstractmethod\n    def terminate_instance(self,instance):\n\n    @abstractmethod\n    def update_instance_info(self,instance):    \n\n# GLOBAL methods \n\ndef get_client(provider=\'aws\',maestro=\'local\'):\n\ndef init_instance_name(instance_config):\n\ndef debug(level,*args,**kwargs):\n```\n\n# CloudSend usage\n\n## Python programmatic use\n\nNote: this demo works the same way, whether CloudSend runs locally or remotely\n\n```python\n\nfrom cloudsend      import provider as cloudsend\nfrom cloudsend.core import CloudSendProcessState\nimport asyncio \n\n# load config\nconfig = __import__(config).config\n\n# create provider: this loads the config\nprovider = cloudsend.get_client(config)\n\n# start the provider: this attempts to create the instances\nawait provider.start()\n\n# deploy the necessary stuff onto the instances\nawait provider.deploy()\n\n# run the jobs and get active processes objects back\nrun_session = await provider.run()\n\n# wait for the active proccesses to be done or aborted:\nawait provider.wait(CloudSendProcessState.DONE|CloudSendProcessState.ABORTED)\n\n# you can get the state of all jobs this way:\nawait provider.get_jobs_states()\n# or get the state for a specific run session:\nawait provider.get_jobs_states(run_session)\n\n# you can print processes summary with:\nawait provider.print_jobs_summary()\n\n# get the results file locally\nawait provider.fetch_results(\'./tmp\')\n\n```\n\n## CLI use\n\nNote: the commands below work the same way, whether CloudSend runs locally or remotely\n\n```bash\nCLI commands go here\n```\n\n# Contributing\nPull requests are welcome. For major changes, please open an issue first to discuss what you would like to change.\n\nPlease make sure to update tests as appropriate.\n\n# License\n[MIT](https://choosealicense.com/licenses/mit/)',
    'author': 'Your Name',
    'author_email': 'you@example.com',
    'maintainer': 'None',
    'maintainer_email': 'None',
    'url': 'https://github.com/benbenz/cloudsend',
    'packages': packages,
    'package_data': package_data,
    'install_requires': install_requires,
    'entry_points': entry_points,
    'python_requires': '>=3.9,<4.0',
}


setup(**setup_kwargs)



import socket , sys
import time
from io import StringIO , TextIOWrapper

HOST = 'localhost' #'127.0.0.1' #'0.0.0.0'
PORT = 5000

def maestro_client(command):
    while True:
        attempts = 0 
        try:
            with socket.socket(socket.AF_INET, socket.SOCK_STREAM) as s:
                s.connect((HOST, PORT))
                #sock_pipe = s.makefile('rw')
                sock_pipe = s.makefile('rw',buffering=1) #TextIOWrapper(s.makefile('rwb',buffering=0), write_through=True)
                #sys.stdout = sock_pipe
                sock_pipe.write(command+"\n")
                sock_pipe.flush()
                try:
                    while True:
                        line = sock_pipe.readline()
                        if not line:
                            break
                        print(line.strip())
                except ConnectionResetError as cre:
                    sock_pipe.readline()
                    print(line.strip())
                    sock_pipe.flush()
                    sock_pipe.close()
                except Exception as e:
                    sock_pipe.flush()
                    sock_pipe.close()   
            break
        except ConnectionRefusedError as err:
            print("waiting on server ...")
            time.sleep(5)
            attempts += 1
            if attempts > 40:
                print("Error: can't connect to server ...")
                sys.exit(99)

def main():
    if len(sys.argv)<2:
        print("python3 -m cloudsend.maestroclient CMD [IP_ADDR]")
        sys.exit()

    if len(sys.argv)>3:
        ip_addr = sys.argv[len(sys.argv)-1]
        command = sys.argv[len(sys.argv)-2]
        HOST    = ip_addr
    else:
        command = sys.argv[len(sys.argv)-1]
        ip_addr = None

    maestro_client(command)  
        
if __name__ == '__main__':
    main()